"""
Smartplug device
"""
import time

from .overview import Overview

OVERVIEW_URL = '/overview/smartplug'
COMMAND_URL = '/smartplugs/onoffplug.cmd'
DETAILS_URL = '/smarthome/{}/details'


class Smartplug(object):
    """ Smartplug device

    Args:
        session (verisure.session): Current session
    """

    def __init__(self, session):
        self._session = session

    def get(self):
        """ Get device overview """
        status = self._session.get(OVERVIEW_URL)
        return [Overview('smartplug', val) for val in status]

    def set(self, device_id, value):
        """ Set device status

        Args:
            device_id (str): Id of the smartplug
            value (str): new status, 'on' or 'off'
        """
        data = {
            'targetDeviceLabel': device_id,
            'targetOn': value
        }
        return not self._session.post(COMMAND_URL, data)

    def get_details(self, device_id):
        """ Get details from a smartplug

        Args:
            device_id (str): Id of the smartplug
        """
        return self._session.get(DETAILS_URL.format(
            device_id.upper().replace(' ', '%20')))

    def set_location(self, device_id, location):
        """ Set smartplug location

        Args:
            device_id (str): Id of the smartplug
            location (str): New location
        """
        details_url = DETAILS_URL.format(
            device_id.upper().replace(' ', '%20'))
        details = self._session.get(details_url)
        details['location'] = location
        self._session.put(details_url, details)

    def wait_while_updating(self, device_id, value, max_request_count=100):
        """ Wait for device status to update

        Args:
            device_id (str): Id of the smartplug
            value (str): status to wait for, 'on' or 'off'
            max_request_count (int): maximum number of post requests

            Returns: retries if success else -1

        """

        for counter in range(max_request_count):
            if [overview for overview in self.get()
                    if overview.id == device_id and overview.status == value]:
                return counter
            time.sleep(1)
        return -1



""" Represents a MyPages session """

import json

import re

import requests

# this import is depending on python version
try:
    import HTMLParser
    UNESCAPE = HTMLParser.HTMLParser().unescape
except ImportError:
    import html
    UNESCAPE = html.unescape


DOMAIN = 'https://mypages.verisure.com'
URL_LOGIN = DOMAIN + '/j_spring_security_check?locale=en_GB'
URL_START = DOMAIN + '/uk/start.html'
RESPONSE_TIMEOUT = 3

CSRF_REGEX = re.compile(
    r'\<input type="hidden" name="_csrf" value="' +
    r'(?P<csrf>(.*?))' +
    r'" /\>')

TITLE_REGEX = re.compile(r'\<title\>(?P<title>(.*))\</title\>')


class Error(Exception):
    ''' mypages error '''
    pass


class LoginError(Error):
    ''' login failed '''
    pass


class ResponseError(Error):
    ''' Unexcpected response '''
    pass


class LoggedOutError(ResponseError):
    ''' Unexcpected response '''
    pass


class TemporarilyUnavailableError(ResponseError):
    ''' My Pages is temporarily unavailable '''
    pass


class MaintenanceError(ResponseError):
    ''' My Pages is currently in maintenance '''
    pass


class Session(object):
    """ Verisure session """

    def __init__(self, username, password):
        self._session = None
        self._username = username
        self._password = password
        self.csrf = ''

    def login(self):
        """ Login to mypages

        Login before calling any read or write commands

        """
        self._session = requests.Session()
        auth = {
            'j_username': self._username,
            'j_password': self._password
        }
        req = requests.Request(
            'POST',
            URL_LOGIN,
            cookies=dict(self._session.cookies),
            data=auth
        ).prepare()
        response = self._session.send(req, timeout=RESPONSE_TIMEOUT)
        self.validate_response(response)
        status = self.json_to_dict(response.text)
        if not status['status'] == 'ok':
            raise LoginError(status['message'])
        self.csrf = self._get_csrf()

    def logout(self):
        """ Ends session

        note:
            Ends the session, will not call the logout command on mypages

        """
        self._session.close()
        self._session = None

    def get(self, url):
        """ Read all statuses of a device type """
        self._ensure_session()
        try:
            response = self._session.get(DOMAIN + url)
        except Exception as e:
            raise Error(e)
        self.validate_response(response)
        return self.json_to_dict(UNESCAPE(response.text))

    def post(self, url, data):
        """ send post request """
        self._ensure_session()
        req = requests.Request(
            'POST',
            DOMAIN + url,
            cookies=dict(self._session.cookies),
            headers={'X-CSRF-TOKEN': self.csrf},
            data=data
        ).prepare()
        response = self._session.send(
            req,
            timeout=RESPONSE_TIMEOUT)
        self.validate_response(response)
        return response.text

    def put(self, url, data):
        """ send put request """
        self._ensure_session()
        req = requests.Request(
            'PUT',
            DOMAIN + url,
            cookies=dict(self._session.cookies),
            headers={
                'X-CSRF-TOKEN': self.csrf,
                'content-type': 'application/json'},
            data=json.dumps(data)
        ).prepare()
        response = self._session.send(
            req,
            timeout=RESPONSE_TIMEOUT)
        self.validate_response(response)
        return response.text

    def _get_csrf(self):
        """ Retreive X-CSRF-TOKEN from start.html """
        response = self._session.get(
            URL_START,
            timeout=RESPONSE_TIMEOUT)
        self.validate_response(response)
        match = CSRF_REGEX.search(response.text)
        return match.group('csrf')

    def _ensure_session(self):
        ''' ensures that a session is created '''
        if not self._session:
            raise Error('Session not started')

    def json_to_dict(self, doc):
        ''' transform json with unicode characters to dict '''
        if not doc:
            return ''
        try:
            return json.loads(doc)
        except ValueError as e:
            self.raise_response_error(
                doc,
                ResponseError(
                    'Unable to convert to JSON, '
                    'Error: {0} - Data: {1}'.format(e, doc)))

    def validate_response(self, response):
        """ Verify that response is OK """
        if response.status_code == 200:
            return
        self.raise_response_error(
            response.text,
            ResponseError(
                'Unable to validate response form My Pages'
                ', status code: {0} - Data: {1}'.format(
                    response.status_code,
                    response.text.encode('utf-8'))))

    @staticmethod
    def raise_response_error(doc, default_error):
        """ Create correct error type from response """
        match = TITLE_REGEX.search(doc)
        if not match:
            raise default_error
        if match == 'My Pages is temporarily unavailable -  Verisure':
            raise TemporarilyUnavailableError('Temporarily unavailable')
        if match == 'My Pages - Maintenance -  Verisure':
            raise MaintenanceError('Maintenance')
        if match == 'Choose country - My Pages - Verisure':
            raise LoggedOutError('Not logged in')
        if match == 'Log in - My Pages - Verisure':
            raise LoggedOutError('Not logged in')
        raise ResponseError(match)



"""
Lock device
"""

from .overview import Overview

OVERVIEW_URL = '/remotecontrol'
COMMAND_URL = '/remotecontrol/lockunlock.cmd'
AUTORELOCK_URL = '/settings/autorelock/'
SETAUTORELOCK_URL = '/settings/setautorelock.cmd'


class Lock(object):
    """ Lock device

    Args:
        session (verisure.session): Current session
    """

    def __init__(self, session):
        self._session = session

    def get(self):
        """ Get device overview """
        status = self._session.get(OVERVIEW_URL)
        return [Overview('lock', val) for val in status
                if val['type'] == 'DOOR_LOCK']

    def set(self, code, device_id, state):
        """ set status of alarm component

            Args:
                code (str): Personal alarm code (four digits)
                device_id (str): lock device id
                state (str): 'LOCKED', or 'UNLOCKED'

        """
        data = {
            'code': code,
            'deviceLabel': device_id,
            'state': state
        }
        return not self._session.post(COMMAND_URL, data)

    def get_autorelock(self):
        """ Get autorelock status """
        return self._session.get(AUTORELOCK_URL)

    def set_autorelock(self, device_id, autorelock):
        """ Set autorelock enabled

            Args:
                device_id (str): lock device id
                autorelock (bool): True to set, else False
        """
        if autorelock:
            data = {
                "_csrf": self._session.csrf,
                "_doorLockDevices[0].autoRelockEnabled": "on",
                "doorLockDevices[0].autoRelockEnabled": "true",
                "enabledDoorLocks": device_id
            }
        else:
            data = {
                "_csrf": self._session.csrf,
                "_doorLockDevices[0].autoRelockEnabled": "on",
                "enabledDoorLocks": ""
            }
        return not self._session.post(SETAUTORELOCK_URL, data)



"""
Overview of device, deynamically created from the device attributes
"""


class Overview(object):
    ''' mypages device overview

        Args:
            overview_type (str): Type of device
            status (str): the json returned from mypages

    '''

    def __init__(self, overview_type, status):
        self.__dict__.update(status)
        self._overview_type = overview_type

    def get_typename(self):
        """ get the type of device

            Returns (str): name of the device type

        """

        return self._overview_type

    def get_status(self):
        ''' get all status items as a dict

            Returns (dict): All status items as a dict

        '''

        return [(key, value) for (key, value)
                in self.__dict__.items() if not key.startswith('_')]
